package net.sf.corn.exception;

import java.io.InputStream;
import java.util.Properties;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.RejectedExecutionException;
import java.util.concurrent.ThreadPoolExecutor;
import java.util.concurrent.TimeUnit;

/**
 * This class is the bus implementation that handle executions of exception events related actions.
 * Although it's designed to work with {@link ExceptionBuilder} and {@link ExceptionCatcher}, you can use its
 * methods for monitoring purposes or changing runtime parameters for performance tuning.
 * As its default parameter it allows only a single action thread can be alive, maximum 10000 threads
 * can be in the pool without a time limit. The bus configuration can be also managed by changing the values of
 * "ExceptionEventsBus.properties" configuration file in the classpath
 * root, so the bus reads that during initialization and sets the parameters.
 * A sample configuration file content :
 * <p><code>
 * &#47;&#47;concurrently executed thread size</br>
 * threadConcurrentSize=2</br>
 * &#47;&#47;Maximum pool size</br>
 * threadMaxPoolSize=20000</br>
 * &#47;&#47;Thread max waiting time in the pool before executing it
 * &#47;&#47;Wait for ever value is 0</br>
 * threadKeepAliveTimeInMS=0</br>
 * &#47;&#47;Enable/disable thread execution timeout</br>
 * &#47;&#47;Enabling it cause keep alive time also valid for thread execution</br>
 * enableThreadExecutionTimeout=false</tr>
 * </code></p>
 * @author Serhat Dirik
 */
public class ExceptionEventsBus {
	public static final String CONFIG_RESOURCE_NAME = "ExceptionEventsBus.properties";
	public static final String CONFIG_KEY_THREAD_CONCURRENT_SIZE = "threadConcurrentSize";
	public static final String CONFIG_KEY_THREAD_MAX_POOL_SIZE = "threadMaxPoolSize";
	public static final String CONFIG_KEY_THREAD_KEEP_ALIVE_TIME_IN_MS = "threadKeepAliveTimeInMS";
	public static final String CONFIG_KEY_TIMEOUT_ENABLED_FOR_THREAD_EXECUTION = "enableThreadExecutionTimeout";
	
	private static ThreadPoolExecutor executor = null;

	static {
		int threadConcurrentSize = 1;
		int threadPoolSize = 10000;
		long keepAliveTimeInMS = 0;
		boolean allowThreadTimeout=false;
		
		ClassLoader loader = Thread.currentThread().getContextClassLoader();
		InputStream configStream = loader
				.getResourceAsStream(CONFIG_RESOURCE_NAME);
		if (configStream != null) {
			Properties props = new Properties();
			try {
				props.load(configStream);
				String prop = props
						.getProperty(CONFIG_KEY_THREAD_MAX_POOL_SIZE);
				threadPoolSize = prop == null ? threadPoolSize : Integer
						.valueOf(prop);
				prop = props.getProperty(CONFIG_KEY_THREAD_CONCURRENT_SIZE);
				threadConcurrentSize = prop == null ? threadConcurrentSize
						: Integer.valueOf(prop);
				prop = props
						.getProperty(CONFIG_KEY_THREAD_KEEP_ALIVE_TIME_IN_MS);
				keepAliveTimeInMS = prop == null ? keepAliveTimeInMS : Long
						.valueOf(prop);
				prop = props
						.getProperty(CONFIG_KEY_TIMEOUT_ENABLED_FOR_THREAD_EXECUTION);
				allowThreadTimeout = prop == null? allowThreadTimeout : Boolean.valueOf(prop);
			} catch (Exception e) {
			}
		}
		executor = new ThreadPoolExecutor(threadConcurrentSize, threadPoolSize,
				keepAliveTimeInMS, TimeUnit.MILLISECONDS,
				new LinkedBlockingQueue<Runnable>());
		executor.allowCoreThreadTimeOut(allowThreadTimeout);
	}

	/**
	 * Returns true if this pool allows core threads to time out and terminate
	 * if no tasks arrive within the keepAlive time, being replaced if needed
	 * when new tasks arrive. When true, the same keep-alive policy applying to
	 * non-core threads applies also to core threads. When false (the default),
	 * core threads are never terminated due to lack of incoming tasks.
	 * 
	 * @return {@code true} if core threads are allowed to time out, else
	 *         {@code false}
	 * 
	 */
	public static boolean allowsCoreThreadTimeOut() {
		return executor.allowsCoreThreadTimeOut();
	}

	/**
	 * Sets the policy governing whether core threads may time out and terminate
	 * if no tasks arrive within the keep-alive time, being replaced if needed
	 * when new tasks arrive. When false, core threads are never terminated due
	 * to lack of incoming tasks. When true, the same keep-alive policy applying
	 * to non-core threads applies also to core threads. To avoid continual
	 * thread replacement, the keep-alive time must be greater than zero when
	 * setting {@code true}. This method should in general be called before the
	 * pool is actively used.
	 * 
	 * @param value
	 *            {@code true} if should time out, else {@code false}
	 * @throws IllegalArgumentException
	 *             if value is {@code true} and the current keep-alive time is
	 *             not greater than zero
	 * 
	 */
	public static void allowCoreThreadTimeOut(boolean value) {
		executor.allowCoreThreadTimeOut(value);
	}

	/**
	 * Sets the core number of threads which specify concurrency on runtime.
	 * This overrides any value set in the constructor. If the new value is
	 * smaller than the current value, excess existing threads will be
	 * terminated when they next become idle. If larger, new threads will, if
	 * needed, be started to execute any queued tasks.
	 * 
	 * @param corePoolSize
	 *            the new core size
	 * @throws IllegalArgumentException
	 *             if {@code corePoolSize < 0}
	 * @see #getCorePoolSize
	 */
	public static void setCorePoolSize(int corePoolSize) {
		executor.setCorePoolSize(corePoolSize);
	}

	/**
	 * Returns the core number of threads.
	 * 
	 * @return the core number of threads
	 * @see #setCorePoolSize
	 */
	public static int getCorePoolSize() {
		return executor.getCorePoolSize();
	}

	/**
	 * Sets the maximum allowed number of threads. This overrides any value set
	 * in the constructor. If the new value is smaller than the current value,
	 * excess existing threads will be terminated when they next become idle.
	 * 
	 * @param maximumPoolSize
	 *            the new maximum
	 * @throws IllegalArgumentException
	 *             if the new maximum is less than or equal to zero, or less
	 *             than the {@linkplain #getCorePoolSize core pool size}
	 * @see #getMaximumPoolSize
	 */
	public static void setMaximumPoolSize(int maximumPoolSize) {
		executor.setMaximumPoolSize(maximumPoolSize);
	}

	/**
	 * Returns the maximum allowed number of threads.
	 * 
	 * @return the maximum allowed number of threads
	 * @see #setMaximumPoolSize
	 */
	public static int getMaximumPoolSize() {
		return executor.getMaximumPoolSize();
	}

	/**
	 * Sets the time limit for which threads may remain idle before being
	 * terminated. If there are more than the core number of threads currently
	 * in the pool, after waiting this amount of time without processing a task,
	 * excess threads will be terminated. This overrides any value set in the
	 * constructor.
	 * 
	 * @param time
	 *            the time to wait. A time value of zero will cause excess
	 *            threads to terminate immediately after executing tasks.
	 * @param unit
	 *            the time unit of the {@code time} argument
	 * @throws IllegalArgumentException
	 *             if {@code time} less than zero or if {@code time} is zero and
	 *             {@code allowsCoreThreadTimeOut}
	 * @see #getKeepAliveTime
	 */
	public static void setKeepAliveTime(long time, TimeUnit unit) {
		executor.setKeepAliveTime(time, unit);
	}

	/**
	 * Returns the thread keep-alive time, which is the amount of time that
	 * threads in excess of the core pool size may remain idle before being
	 * terminated.
	 * 
	 * @param unit
	 *            the desired time unit of the result
	 * @return the time limit
	 */
	public static long getKeepAliveTime(TimeUnit unit) {
		return executor.getKeepAliveTime(unit);
	}

	/**
	 * Returns the current number of threads in the pool.
	 * 
	 * @return the number of threads
	 */
	public static int getPoolSize() {
		return executor.getPoolSize();
	}

	/**
	 * Returns the approximate number of threads that are actively executing
	 * tasks.
	 * 
	 * @return the number of threads
	 */
	public static int getActiveCount() {
		return executor.getActiveCount();
	}

	/**
	 * Returns the approximate total number of tasks that have ever been
	 * scheduled for execution. Because the states of tasks and threads may
	 * change dynamically during computation, the returned value is only an
	 * approximation.
	 * 
	 * @return the number of tasks
	 */
	public static long getTaskCount() {
		return executor.getTaskCount();
	}

	/**
	 * Returns the approximate total number of tasks that have completed
	 * execution. Because the states of tasks and threads may change dynamically
	 * during computation, the returned value is only an approximation, but one
	 * that does not ever decrease across successive calls.
	 * 
	 * @return the number of tasks
	 */
	public static long getCompletedTaskCount() {
		return executor.getCompletedTaskCount();
	}

	/**
	 * Returns the approximate total number of waiting tasks that waiting for
	 * execution. Because the states of tasks and threads may change dynamically
	 * during computation, the returned value is only an approximation.
	 * 
	 * @return the number of waiting tasks
	 */
	public static long getWaitingTaskCount() {
		return executor.getTaskCount() - executor.getCompletedTaskCount();
	}

	/**
	 * Executes the given task sometime in the future. The task may execute in a
	 * new thread or in an existing pooled thread.
	 * 
	 * If the task cannot be submitted for execution, either because this
	 * executor has been shutdown or because its capacity has been reached, the
	 * task is handled by the current {@code RejectedExecutionHandler}.
	 * 
	 * @param command
	 *            the task to execute
	 * @throws RejectedExecutionException
	 *             at discretion of {@code RejectedExecutionHandler}, if the
	 *             task cannot be accepted for execution
	 * @throws NullPointerException
	 *             if {@code command} is null
	 */
	static void execute(Runnable command) {
		if(command != null)	executor.execute(command);
	}

}
